"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const stub_1 = __importDefault(require("./stub"));
const debug_1 = __importDefault(require("debug"));
const function_table_1 = __importDefault(require("./function-table"));
const dlopen_1 = __importDefault(require("./dlopen"));
const global_offset_table_1 = __importDefault(require("./global-offset-table"));
const log = (0, debug_1.default)("dylink");
const logImport = (0, debug_1.default)("dylink:import");
async function importWebAssemblyDlopen({ path, importObject, importWebAssembly, importWebAssemblySync, readFileSync, stub, allowMainExports, }) {
    let mainInstance = null;
    if (importObject == null) {
        importObject = {};
    }
    let { env } = importObject;
    if (env == null) {
        env = importObject.env = {};
    }
    let { memory } = env;
    if (memory == null) {
        memory = env.memory = new WebAssembly.Memory({ initial: 10 });
    }
    let { __indirect_function_table } = env;
    if (__indirect_function_table == null) {
        // TODO: Make the 1000 bigger if your main module has a large number of function pointers
        // Maybe we need to parse the wasm bundle in general (that's what emscripten does).
        // Note that this is only potentially an issue for the main core WASM module, not the
        // dynamic libraries that get loaded at runtime.
        __indirect_function_table = env.__indirect_function_table =
            new WebAssembly.Table({ initial: 1500, element: "anyfunc" });
    }
    const functionTable = new function_table_1.default(__indirect_function_table);
    function functionViaPointer(key) {
        if (mainInstance == null)
            return; // not yet available
        const f = mainInstance.exports[`__WASM_EXPORT__${key}`];
        if (f == null)
            return;
        const ptr = f();
        log("functionViaPointer", key, ptr);
        return functionTable.get(ptr);
    }
    function getFunction(name, path = "") {
        log("getFunction", name);
        let f = importObject?.env?.[name];
        if (f != null) {
            log("getFunction ", name, "from env");
            return f;
        }
        f = functionViaPointer(name);
        if (f != null) {
            log("getFunction ", name, "from function pointer");
            return f;
        }
        f = dlopenManager.getFunction(name);
        if (f != null) {
            log("getFunction ", name, "from other library");
            return f;
        }
        if (allowMainExports) {
            /*
            Any other way of resolving a function needed in a dynamic import that isn't
            a function pointer is NOT going to work in general:
            It will segfault or be 1000x too slow.  Every function
            needs to be via a pointer. The following doesn't work *in general*.  In addition to
            speed, there are C functions that make no sense to call via WASM,
            since they have signatures that are more complicated than WASM supports.
            */
            f = mainInstance?.exports[name];
            if (f != null) {
                log("getFunction ", name, "from mainInstance exports (potentially dangerous!)");
                return f;
            }
        }
        // **TODO: this is a temporary whitelist for some mangled C++ symbols in the numpy build**
        if (path?.includes("numpy") && name.startsWith("_Z")) {
            return () => {
                console.log("WARNING: calling dangerous stub for ", name);
            };
        }
        if (path) {
            // this is a dynamic library import, so fail at this point:
            throw Error(`${name} -- undefined when importing ${path}`);
        }
        return importObjectWithPossibleStub.env[name];
    }
    function getMainInstance() {
        if (mainInstance == null)
            throw Error("bug");
        return mainInstance;
    }
    function getMainInstanceExports() {
        if (mainInstance?.exports == null)
            throw Error("bug");
        return mainInstance.exports;
    }
    const globalOffsetTable = new global_offset_table_1.default(getMainInstanceExports, functionTable);
    const dlopenManager = new dlopen_1.default(getFunction, memory, globalOffsetTable, functionTable, readFileSync, importObject, importWebAssemblySync, getMainInstanceExports, getMainInstance);
    dlopenManager.add_dlmethods(env);
    const importObjectWithPossibleStub = stub
        ? {
            ...importObject,
            env: (0, stub_1.default)(importObject.env, functionViaPointer, stub),
        }
        : importObject;
    let t0 = 0;
    if (logImport.enabled) {
        t0 = new Date().valueOf();
        logImport("importing ", path);
    }
    mainInstance =
        importWebAssembly != null
            ? await importWebAssembly(path, importObjectWithPossibleStub)
            : importWebAssemblySync(path, importObjectWithPossibleStub);
    if (logImport.enabled) {
        logImport("imported ", path, ", time =", new Date().valueOf() - t0, "ms");
    }
    if (mainInstance.exports.__wasm_call_ctors != null) {
        // We also **MUST** explicitly call the WASM constructors. This is
        // a library function that is part of the zig libc code.  We have
        // to call this because the wasm file is built using build-lib, so
        // there is no main that does this.  This call does things like
        // setup the filesystem mapping.    Yes, it took me **days**
        // to figure this out, including reading a lot of assembly code. :shrug:
        mainInstance.exports.__wasm_call_ctors();
    }
    functionTable.updateAfterImport();
    // TODO
    mainInstance.env = env;
    mainInstance.getDlopenState = () => {
        return {
            dlopen: dlopenManager.getState(),
            got: globalOffsetTable.getState(),
        };
    };
    mainInstance.setDlopenState = (state) => {
        const { dlopen, got } = state;
        dlopenManager.setState(dlopen);
        globalOffsetTable.setState(got);
    };
    return mainInstance;
}
exports.default = importWebAssemblyDlopen;
//# sourceMappingURL=import.js.map